import unittest

import numpy as np
import Levenshtein
from scipy.stats import norm

from peyes._utils.metric_utils import transition_matrix, dprime_and_criterion, _dprime_rates
from peyes._utils.metric_utils import normalized_levenshtein_distance as nld


class TestMetricUtils(unittest.TestCase):

    def test_transition_matrix(self):
        seq = [0, 1, 2, 3, 2, 3, 2, 2]
        self.assertTrue(
            np.array_equal(transition_matrix(seq).values,
                           np.array([[1, 0, 0], [0, 1, 0], [0, 1, 2], [0, 2, 0]]))
        )
        self.assertTrue(
            np.array_equal(transition_matrix(seq, True).values,
                           np.array([[1, 0, 0], [0, 1, 0], [0, 1/3, 2/3], [0, 1, 0]]))
        )

    def test_nld(self):
        gt = "kitten"
        pred = "sitting"
        self.assertEqual(Levenshtein.distance(gt, pred) / len(gt), nld(gt, pred))

    def test_dprime_rates(self):
        p, n, pp, tp = 10, 20, 15, 5
        # hr, far = dprime_rates(p, n, pp, tp, None)
        self.assertEqual((tp/p, (pp-tp)/n), _dprime_rates(p, n, pp, tp, None))
        self.assertEqual((tp/p, (pp-tp)/n), _dprime_rates(p, n, pp, tp, "macmillan"))
        self.assertEqual((tp/p, (pp-tp)/n), _dprime_rates(p, n, pp, tp, "loglinear"))
        self.assertEqual((tp/p, (pp-tp)/n), _dprime_rates(p, n, pp, tp, "foo"))
        tp = 10
        self.assertEqual((tp/p, (pp-tp)/n), _dprime_rates(p, n, pp, tp, None))
        self.assertEqual((1-0.5/p, (pp-tp)/n), _dprime_rates(p, n, pp, tp, "macmillan"))
        self.assertTrue(np.isclose(
            (tp + p / (p+n)) / (p + 2 * p / (p+n)), _dprime_rates(p, n, pp, tp, "loglinear")[0]
        ))
        self.assertTrue(np.isclose(
            (pp-tp + 1 - p / (p+n)) / (n + 2 * (1 - p / (p+n))), _dprime_rates(p, n, pp, tp, "loglinear")[1]
        ))
        self.assertRaises(ValueError, _dprime_rates, p, n, pp, tp, "foo")
        tp = 15
        self.assertRaises(AssertionError, _dprime_rates, p, n, pp, tp, None)
        p = tp = 0
        self.assertTrue(np.isnan(_dprime_rates(p, n, pp, tp, None)[0]))

    def test_dprime(self):
        p = n = pp = 10
        tp = 5
        exp = norm.ppf(tp / p) - norm.ppf((pp - tp) / n)
        self.assertEqual(dprime_and_criterion(p, n, pp, tp, None)[0], exp)
        self.assertEqual(dprime_and_criterion(p, n, pp, tp, "macmillan")[0], exp)
        self.assertEqual(dprime_and_criterion(p, n, pp, tp, "loglinear")[0], exp)
        self.assertEqual(dprime_and_criterion(p, n, pp, tp, "foo")[0], exp)
        tp = 10
        self.assertTrue(np.isinf(dprime_and_criterion(p, n, pp, tp, None)[0]))
        self.assertEqual(
            dprime_and_criterion(p, n, pp, tp, "macmillan")[0], norm.ppf(1 - 0.5 / p) - norm.ppf(0.5 / n)
        )
        hr_ll, far_ll = _dprime_rates(p, n, pp, tp, "loglinear")
        self.assertEqual(
            dprime_and_criterion(p, n, pp, tp, "loglinear")[0], norm.ppf(hr_ll) - norm.ppf(far_ll)
        )
        self.assertRaises(ValueError, dprime_and_criterion, p, n, pp, tp, "foo")

    def test_criterion(self):
        p = n = pp = 10
        tp = 5
        exp = -0.5 * (norm.ppf(tp / p) + norm.ppf((pp - tp) / n))
        self.assertEqual(dprime_and_criterion(p, n, pp, tp, None)[1], exp)
        self.assertEqual(dprime_and_criterion(p, n, pp, tp, "macmillan")[1], exp)
        self.assertEqual(dprime_and_criterion(p, n, pp, tp, "loglinear")[1], exp)
        self.assertEqual(dprime_and_criterion(p, n, pp, tp, "foo")[1], exp)
        tp = 10
        self.assertTrue(np.isnan(dprime_and_criterion(p, n, pp, tp, None)[1]))
        self.assertEqual(
            dprime_and_criterion(p, n, pp, tp, "macmillan")[1], -0.5 * (norm.ppf(1 - 0.5 / p) + norm.ppf(0.5 / n))
        )
        hr_ll, far_ll = _dprime_rates(p, n, pp, tp, "loglinear")
        self.assertEqual(
            dprime_and_criterion(p, n, pp, tp, "loglinear")[1], -0.5 * (norm.ppf(hr_ll) + norm.ppf(far_ll))
        )
        self.assertRaises(ValueError, dprime_and_criterion, p, n, pp, tp, "foo")




import warnings
from typing import Tuple

import numpy as np

import peyes._utils.constants as cnst


def cast_to_integers(xs: np.ndarray, ys: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """
    Casts the given x and y coordinates to integers, rounding down to the nearest smaller integer.
    :param xs: 1D array of x coordinates
    :param ys: 1D array of y coordinates

    :return: 1D arrays of x and y coordinates, cast to integers
    """
    def cast_finites_to_integers(arr: np.ndarray) -> np.ndarray:
        new_arr = arr.copy()
        is_finite = np.isfinite(arr)
        new_arr[is_finite] = new_arr[is_finite].astype(int)
        return new_arr

    x_int = cast_finites_to_integers(xs)
    y_int = cast_finites_to_integers(ys)
    return x_int, y_int


def line_dispersion(xs: np.ndarray, ys: np.ndarray) -> float:
    """
    Calculates the line-dispersion (1D) of the given array of pixels, in pixel units.
    The line-dispersion is the sum of the differences between the maximum and minimum values of the x and y coordinates.

    :param xs: 1D array of x coordinates
    :param ys: 1D array of y coordinates
    :return: sum of dispersion along x and y axes
    """
    x_disp = np.nanmax(xs) - np.nanmin(xs)
    y_disp = np.nanmax(ys) - np.nanmin(ys)
    return x_disp + y_disp


def ellipse_dispersion(xs: np.ndarray, ys: np.ndarray) -> float:
    """
    Calculates the ellipse-dispersion (2D) of the given array of pixels, in pixel^2 units.
    The ellipse-dispersion is the area of the ellipse that has the same horizontal and vertical radii as the
    bounding box of the given pixels.

    :param xs: 1D array of x coordinates
    :param ys: 1D array of y coordinates
    :return: area of ellipse that bounds the given pixels
    """
    horizontal_radius = 0.5 * (np.nanmax(xs) - np.nanmin(xs))
    vertical_radius = 0.5 * (np.nanmax(ys) - np.nanmin(ys))
    return np.pi * horizontal_radius * vertical_radius


def calculate_pixel_size(width: float, height: float, resolution: Tuple[int, int]) -> float:
    """ Calculates the approximate length of one pixel in centimeters (assuming square pixels): cm/px """
    diagonal_length = np.sqrt(  # size of diagonal in centimeters
        np.power(width, 2) + np.power(height, 2)
    )
    diagonal_pixels = np.sqrt(  # size of diagonal in pixels
        np.power(round(resolution[0]), 2) + np.power(round(resolution[1]), 2)
    )
    return diagonal_length / diagonal_pixels


def calculate_velocities(xs: np.ndarray, ys: np.ndarray, ts: np.ndarray) -> np.ndarray:
    """
    Calculates the velocity between subsequent pixels in the given x and y coordinates, in pixels per second.
    :param xs: 1D array of x coordinates
    :param ys: 1D array of y coordinates
    :param ts: 1D array of timestamps (in milliseconds)
    :return: velocity (pixel / second) between subsequent pixels
    """
    assert len(xs) == len(ys) == len(ts), "`xs`, `ys` and `ts` arrays must be of the same length"
    time_diff = np.diff(ts) / cnst.MILLISECONDS_PER_SECOND  # convert from milliseconds to seconds
    px_distance = np.sqrt(np.power(np.diff(xs), 2) + np.power(np.diff(ys), 2))
    velocities = px_distance / time_diff
    velocities = np.concatenate(([np.nan], velocities))  # first velocity is NaN
    return velocities


def pixels_to_visual_angle(num_px: float, d: float, pixel_size: float, use_radians=False) -> float:
    """
    Calculates the visual angle that corresponds to `num_px` pixels, given that the viewer is sitting at a distance of
    `d` centimeters from the screen, and that the size of each pixel is `pixel_size` centimeters.

    See details on calculations in Kaiser, Peter K. "Calculation of Visual Angle". The Joy of Visual Perception: A Web Book:
        http://www.yorku.ca/eye/visangle.htm

    :param num_px: the number of pixels.
    :param d: the distance (in cm) from the screen.
    :param pixel_size: the size (of the diagonal) of a pixel (in cm).
    :param use_radians: if True, returns the angle in radians. Otherwise, returns the angle in degrees.

    :return: the visual angle (in degrees) that corresponds to the given number of pixels.
        If `num_px` is not finite, returns np.nan.

    :raises ValueError: if any of the arguments is negative.
    """
    if not np.isfinite([num_px, d, pixel_size]).all():
        return np.nan
    if (np.array([num_px, d, pixel_size]) < 0).any():
        raise ValueError("arguments `num_px`, `d` and `pixel_size` must be non-negative numbers")
    full_edge_cm = num_px * pixel_size
    half_angle = np.arctan(full_edge_cm / (2 * d))
    angle = 2 * half_angle
    if use_radians:
        return angle
    return np.rad2deg(angle)


def visual_angle_to_pixels(
        angle: float, d: float, pixel_size: float, use_radians: bool = False, keep_sign: bool = False
) -> float:
    """
    Calculates the number of pixels that are equivalent to a visual angle `angle`, given that the viewer is sitting at
    a distance of `d` centimeters from the screen, and that the size of each pixel is `pixel_size` centimeters.
    See details on calculations in Kaiser, Peter K. "Calculation of Visual Angle". The Joy of Visual Perception: A Web
        Book: http://www.yorku.ca/eye/visangle.htm
    :param angle: the visual angle (degrees/radians).
    :param d: the distance of the viewer from the screen (cm).
    :param pixel_size: the size of each pixel (cm).
    :param use_radians: if True, `angle` is in radians; otherwise, it is in degrees.
    :param keep_sign: if True, returns a negative number if `angle` is negative. Otherwise, returns the absolute value.
    :return: the number of pixels that correspond to the given visual angle. If `angle` is not finite, returns np.nan.
    """
    if not np.isfinite([angle, d, pixel_size]).all():
        return np.nan
    if (np.array([d, pixel_size]) <= 0).any():
        raise ValueError("Arguments `d` and `pixel_size` must be positive numbers")
    if angle == 0:
        return 0
    abs_angle = abs(angle) if use_radians else np.deg2rad(abs(angle))
    half_edge_cm = d * np.tan(abs_angle / 2)        # half the edge size in cm
    edge_pixels = 2 * half_edge_cm / pixel_size     # full edge size in pixels
    edge_pixels = np.sign(angle) * edge_pixels if keep_sign else edge_pixels
    return edge_pixels


def calculate_azimuth(
        p1: Tuple[float, float], p2: Tuple[float, float], zero_direction: str = 'E', use_radians: bool = False
) -> float:
    """
    Calculates the counter-clockwise angle between the line starting from p1 and ending at p2, and the line starting
    from p1 and pointing in the direction of `zero_direction`.

    :param p1: the (x,y) coordinates of the starting point of the line
    :param p2: the (x,y) coordinates of the ending point of the line
    :param zero_direction: the direction of the zero angle. Must be one of 'E', 'W', 'S', 'N' (case insensitive).
    :param use_radians: if True, returns the angle in radians. Otherwise, returns the angle in degrees.

    :return: the angle between the two lines, in range [0, 2*pi) or [0, 360), or np.nan if either p1 or p2 is invalid.

    :raises ValueError: if zero_direction is not one of 'E', 'W', 'S', 'N' (case insensitive).
    """
    # exit early for invalid pixels:
    if not (np.isfinite(p1[0]) and np.isfinite(p1[1]) and np.isfinite(p2[0]) and np.isfinite(p2[1])):
        return np.nan
    # verify 0-direction
    zero_direction = zero_direction.upper()
    valid_directions = ['E', 'W', 'S', 'N']
    if zero_direction not in valid_directions:
        raise ValueError(f"zero_direction must be one of {valid_directions}")
    # calc angle & adjust to the desired zero direction
    x1, y1 = p1
    x2, y2 = p2
    angle_rad = np.arctan2(y1 - y2, x2 - x1)  # counter-clockwise angle between line (p1, p2) and right-facing x-axis
    if zero_direction == 'W':
        angle_rad += np.pi
    elif zero_direction == 'S':
        angle_rad += np.pi / 2
    elif zero_direction == 'N':
        angle_rad -= np.pi / 2
    # make sure the angle is in range [0, 2*pi), and return
    angle_rad = angle_rad % (2 * np.pi)
    if use_radians:
        return angle_rad
    return np.rad2deg(angle_rad)



import peyes._utils.constants as cnst
import peyes._DataModels.config as cnfg
from peyes._utils.pixel_utils import calculate_pixel_size as calc_ps
from peyes._utils.event_utils import parse_label
from peyes._DataModels.UnparsedEventLabel import UnparsedEventLabelType


def set_viewer_distance(distance_cm: float) -> None:
    """ Sets the default viewer distance (in centimeters). """
    assert distance_cm > 0, "Viewer distance must be positive"
    cnfg.VIEWER_DISTANCE = distance_cm


def set_screen_monitor(
        width_cm: float = None, height_cm: float = None, width_px: int = None, height_px: int = None,
) -> None:
    """
    Sets the default screen monitor configuration: width and height in centimeters, and resolution in pixels.
    Then, calculates the default pixel size based on the provided dimensions and resolution.
    """
    if width_cm is not None:
        assert width_cm > 0, "Screen width (cm) must be positive"
        cnfg.SCREEN_MONITOR[cnst.WIDTH_STR] = width_cm
    if height_cm is not None:
        assert height_cm > 0, "Screen height (cm) must be positive"
        cnfg.SCREEN_MONITOR[cnst.HEIGHT_STR] = height_cm
    if width_px is not None or height_px is not None:
        width_px = width_px if width_px is not None else cnfg.SCREEN_MONITOR[cnst.RESOLUTION_STR][0]
        height_px = height_px if height_px is not None else cnfg.SCREEN_MONITOR[cnst.RESOLUTION_STR][1]
        assert width_px > 0 and height_px > 0, "Screen resolution (px) must be positive"
        cnfg.SCREEN_MONITOR[cnst.RESOLUTION_STR] = (width_px, height_px)
    cnfg.SCREEN_MONITOR[cnst.PIXEL_SIZE_STR] = calc_ps(
        cnfg.SCREEN_MONITOR[cnst.WIDTH_STR],
        cnfg.SCREEN_MONITOR[cnst.HEIGHT_STR],
        cnfg.SCREEN_MONITOR[cnst.RESOLUTION_STR],
    )


def set_event_configurations(
        event_type: UnparsedEventLabelType,
        min_duration: float = None,
        max_duration: float = None,
        hex_color: str = None,
) -> None:
    """
    Sets the event-type's default configuration:
    - minimum duration (in milliseconds)
    - maximum duration (in milliseconds)
    - plotting color (in hex format)
    """
    if min_duration is not None:
        _set_min_duration(event_type, min_duration)
    if max_duration is not None:
        _set_max_duration(event_type, max_duration)
    if hex_color is not None:
        set_event_color(event_type, hex_color)


def _set_min_duration(event_type: UnparsedEventLabelType, duration: float) -> None:
    """ Sets the minimum duration (in milliseconds) for the specified event type. """
    assert duration > 0, "Minimum duration must be positive"
    label = parse_label(event_type, safe=False)
    cnfg.EVENT_MAPPING[label][cnst.MIN_DURATION_STR] = duration


def _set_max_duration(event_type: UnparsedEventLabelType, duration: float) -> None:
    """ Sets the maximum duration (in milliseconds) for the specified event type. """
    assert duration > 0, "Maximum duration must be positive"
    label = parse_label(event_type, safe=False)
    cnfg.EVENT_MAPPING[label][cnst.MAX_DURATION_STR] = duration


def set_event_color(event_type: UnparsedEventLabelType, hex_color: str) -> None:
    """ Sets the color (in hex format) for the specified event type. """
    assertion_message = "Color must be a valid hex string."
    if len(hex_color) == 7:
        assert hex_color[0] == "#" and all(c in "0123456789abcdefABCDEF" for c in hex_color[1:]), assertion_message
    elif len(hex_color) == 6:
        assert all(c in "0123456789abcdefABCDEF" for c in hex_color), assertion_message
        hex_color = "#" + hex_color
    else:
        raise AssertionError(assertion_message)
    label = parse_label(event_type, safe=False)
    cnfg.EVENT_MAPPING[label][cnst.COLOR_STR] = hex_color




import unittest

import numpy as np

import peyes._utils.constants as cnst
from peyes._utils.pixel_utils import *


class TestPixelUtils(unittest.TestCase):
    D = 1  # distance from screen to eye
    PS = 1  # pixel size

    def test_cast_to_integers(self):
        xs = np.arange(0.1, 5.1, 1)
        ys = np.arange(5.1, 0.1, -1)
        self.assertTrue(
            np.allclose(np.vstack((xs, ys)).astype(int), np.vstack(cast_to_integers(xs, ys)))
        )
        self.assertTrue(
            np.allclose(np.vstack((xs, ys)).astype(int), np.vstack(cast_to_integers(xs, ys)))
        )
        xs[2] = np.nan
        self.assertTrue(
            np.allclose(np.vstack(cast_to_integers(xs, ys)), np.vstack(cast_to_integers(xs, ys)), equal_nan=True)
        )

    def test_line_dispersion(self):
        xs = np.arange(0, 5).astype(float)
        ys = np.arange(5, 0, -1).astype(float)
        self.assertEqual(8, line_dispersion(xs, ys))
        xs[2] = np.nan
        self.assertEqual(8, line_dispersion(xs, ys))

    def test_ellipse_dispersion(self):
        xs = np.arange(0, 5).astype(float)
        ys = np.arange(5, 0, -1).astype(float)
        self.assertTrue(np.isclose(np.pi * 4, ellipse_dispersion(xs, ys)))
        xs[2] = np.nan
        self.assertTrue(np.isclose(np.pi * 4, ellipse_dispersion(xs, ys)))

    def test_calculate_pixel_size(self):
        self.assertEqual(1, calculate_pixel_size(width=1, height=1, resolution=(1, 1)))
        self.assertEqual(0.5, calculate_pixel_size(width=1, height=1, resolution=(2, 2)))
        self.assertTrue(np.isclose(
            0.027844,
            calculate_pixel_size(width=cnst.TOBII_WIDTH, height=cnst.TOBII_HEIGHT, resolution=cnst.TOBII_RESOLUTION)
        ))

    def test_calculate_velocities(self):
        sqrt2 = np.sqrt(2)
        x_coords = np.arange(0, 5)
        y_coords1 = np.arange(0, 5)
        t_coords1 = np.arange(5)
        self.assertTrue(np.allclose(
            np.array([np.nan, sqrt2, sqrt2, sqrt2, sqrt2]) * cnst.MILLISECONDS_PER_SECOND,
            calculate_velocities(x_coords, y_coords1, t_coords1),
            equal_nan=True
        ))
        y_coords2 = np.zeros_like(x_coords)
        self.assertTrue(np.allclose(
            np.array([np.nan, 1, 1, 1, 1]) * cnst.MILLISECONDS_PER_SECOND,
            calculate_velocities(x_coords, y_coords2, t_coords1),
            equal_nan=True
        ))
        t_coords2 = t_coords1 * 2
        self.assertTrue(np.allclose(
            np.array([np.nan, 0.5, 0.5, 0.5, 0.5]) * cnst.MILLISECONDS_PER_SECOND,
            calculate_velocities(x_coords, y_coords2, t_coords2),
            equal_nan=True
        ))
        y_coords3 = y_coords1.copy().astype(float)
        y_coords3[2] = np.nan

        print(calculate_velocities(x_coords, y_coords3, t_coords1))

        self.assertTrue(np.allclose(
            np.array([np.nan, sqrt2, np.nan, np.nan, sqrt2]) * cnst.MILLISECONDS_PER_SECOND,
            calculate_velocities(x_coords, y_coords3, t_coords1),
            equal_nan=True
        ))
        t_coords3 = t_coords1[:-1].copy()
        self.assertRaises(AssertionError, calculate_velocities, x_coords, y_coords1, t_coords3)

    def test_pixels_to_visual_angle(self):
        self.assertEqual(0, pixels_to_visual_angle(num_px=0, d=1, pixel_size=1))
        self.assertEqual(90, pixels_to_visual_angle(num_px=2, d=1, pixel_size=1))
        self.assertEqual(np.pi / 2, pixels_to_visual_angle(num_px=2, d=1, pixel_size=1, use_radians=True))
        self.assertTrue(np.isclose(60, pixels_to_visual_angle(num_px=2*np.sqrt(1/3), d=1, pixel_size=1)))
        self.assertTrue(np.isclose(
            np.pi * 2/3, pixels_to_visual_angle(num_px=2*np.sqrt(3), d=1, pixel_size=1, use_radians=True))
        )
        self.assertTrue(np.isnan(pixels_to_visual_angle(num_px=np.inf, d=1, pixel_size=1)))
        self.assertRaises(ValueError, pixels_to_visual_angle, num_px=-1, d=1, pixel_size=1)
        self.assertRaises(ValueError, pixels_to_visual_angle, num_px=1, d=-1, pixel_size=1)
        self.assertRaises(ValueError, pixels_to_visual_angle, num_px=1, d=1, pixel_size=-1)

    def test_visual_angle_to_pixels(self):
        self.assertEqual(0, visual_angle_to_pixels(angle=0, d=1, pixel_size=1))
        self.assertTrue(np.isclose(2, visual_angle_to_pixels(angle=90, d=1, pixel_size=1)))
        self.assertTrue(np.isclose(2, visual_angle_to_pixels(angle=np.pi / 2, d=1, pixel_size=1, use_radians=True)))
        self.assertTrue(np.isclose(2*np.sqrt(1/3), visual_angle_to_pixels(angle=60, d=1, pixel_size=1)))
        self.assertTrue(np.isclose(
            -2*np.sqrt(3), visual_angle_to_pixels(
                angle=-np.pi * 2/3, d=1, pixel_size=1, use_radians=True, keep_sign=True
            ))
        )
        self.assertTrue(np.isnan(visual_angle_to_pixels(angle=np.inf, d=1, pixel_size=1)))
        self.assertRaises(ValueError, visual_angle_to_pixels, angle=1, d=-1, pixel_size=1)
        self.assertRaises(ValueError, visual_angle_to_pixels, angle=1, d=1, pixel_size=-1)

    def test_calculate_azimuth(self):
        # angles are counter-clockwise from the positive x-axis, with y-axis pointing down
        self.assertEqual(0, calculate_azimuth(p1=(0, 0), p2=(0, 0), use_radians=False))
        self.assertEqual(45, calculate_azimuth(p1=(0, 0), p2=(1, -1), use_radians=False))
        self.assertEqual(315, calculate_azimuth(p1=(0, 0), p2=(1, 1), use_radians=False))
        self.assertEqual(90, calculate_azimuth(p1=(1, 1), p2=(1, -1), use_radians=False))
        self.assertEqual(180, calculate_azimuth(p1=(1, 1), p2=(1, -1), use_radians=False, zero_direction='S'))
        self.assertEqual(np.pi * 3 / 4, calculate_azimuth(p1=(1, 1), p2=(-2, -2), use_radians=True))
        self.assertEqual(np.pi, calculate_azimuth(p1=(1, 0), p2=(-1, 0), use_radians=True))
        self.assertEqual(np.pi * 5 / 4, calculate_azimuth(p1=(1, 0), p2=(0, 1), use_radians=True))
        self.assertEqual(np.pi * 3 / 4, calculate_azimuth(p1=(1, 0), p2=(0, 1), use_radians=True, zero_direction='N'))



import os
from typing import List, Optional, Union

import numpy as np
import pandas as pd
from tqdm import tqdm

import peyes
from peyes._DataModels.UnparsedEventLabel import UnparsedEventLabelType, UnparsedEventLabelSequenceType

import analysis.utils as u
import analysis.process._helpers as h

_MATCHED_FEATURE_NAMES = [
    f"{peyes.constants.ONSET_STR}_{peyes.constants.DIFFERENCE_STR}",
    f"{peyes.constants.OFFSET_STR}_{peyes.constants.DIFFERENCE_STR}",
    f"{peyes.constants.DURATION_STR}_{peyes.constants.DIFFERENCE_STR}",
    f"{peyes.constants.AMPLITUDE_STR}_{peyes.constants.DIFFERENCE_STR}",
    f"{peyes.constants.AZIMUTH_STR}_{peyes.constants.DIFFERENCE_STR}",
    f"center_{peyes.constants.PIXEL_STR}_{peyes.constants.DISTANCE_STR}",
    f"{peyes.constants.TIME_STR}_overlap",
    f"{peyes.constants.TIME_STR}_iou",
    f"{peyes.constants.TIME_STR}_l2",
]
_MATCH_SDT_METRICS = [
    peyes.constants.MATCH_RATIO_STR,
    peyes.constants.PRECISION_STR, peyes.constants.RECALL_STR, peyes.constants.F1_STR,
    peyes.constants.D_PRIME_STR, peyes.constants.CRITERION_STR
]


def run_default(
        dataset_name: str,
        pos_labels: Optional[Union[UnparsedEventLabelType, UnparsedEventLabelSequenceType]] = None,
) -> (pd.DataFrame, pd.DataFrame):
    default_output_dir = h.get_default_output_dir(dataset_name)
    try:
        events = pd.read_pickle(os.path.join(default_output_dir, f"{peyes.constants.EVENTS_STR}.pkl"))
    except FileNotFoundError:
        raise FileNotFoundError(
            f"Couldn't find `{peyes.constants.EVENTS_STR}.pkl` in {default_output_dir}. Please preprocess the dataset first."
        )
    try:
        matches = pd.read_pickle(os.path.join(default_output_dir, f"{peyes.constants.MATCHES_STR}.pkl"))
    except FileNotFoundError:
        raise FileNotFoundError(
            f"Couldn't find `{peyes.constants.MATCHES_STR}.pkl` in {default_output_dir}. Please preprocess the dataset first."
        )
    matches_metrics_dir = os.path.join(default_output_dir, f"{peyes.constants.MATCHES_STR}_{peyes.constants.METRICS_STR}")
    os.makedirs(matches_metrics_dir, exist_ok=True)
    features_fullpath = os.path.join(
        matches_metrics_dir, u.get_filename_for_labels(labels=None, suffix="matched_features", extension="pkl")
    )
    try:
        matched_features = pd.read_pickle(features_fullpath)
    except FileNotFoundError:
        matched_features = calculate_matched_features(matches, features=_MATCHED_FEATURE_NAMES)
        matched_features.to_pickle(features_fullpath)
    sdt_fullpath = os.path.join(
        matches_metrics_dir, u.get_filename_for_labels(pos_labels, suffix="sdt_metrics", extension="pkl")
    )
    try:
        sdt_metrics = pd.read_pickle(sdt_fullpath)
    except FileNotFoundError:
        sdt_metrics = calculate_event_sdt_measures(events, matches, pos_labels)
        sdt_metrics.to_pickle(sdt_fullpath)
    return matched_features, sdt_metrics


def calculate_matched_features(
        matches: pd.DataFrame,
        features: List[str] = None,
):
    if features is None:
        features = _MATCHED_FEATURE_NAMES
    features = [f.lower().strip().replace(" ", "_").replace("-", "_") for f in features]
    if not set(features).issubset(_MATCHED_FEATURE_NAMES):
        raise ValueError(f"Unknown feature(s): {set(features) - set(_MATCHED_FEATURE_NAMES)}")
    results = dict()
    trials = matches.columns.get_level_values(level=peyes.constants.TRIAL_ID_STR).unique()
    gt_labelers = matches.columns.get_level_values(u.GT_STR).unique()
    pred_labelers = matches.columns.get_level_values(u.PRED_STR).unique()
    iterations = matches.columns.get_level_values(peyes.constants.ITERATION_STR).unique()
    matching_schemes = matches.index.get_level_values(u.MATCHING_SCHEME_STR).unique()
    for tr in tqdm(trials, desc="Matched Events :: Features"):
        for gt_labeler in gt_labelers:
            for pred_labeler in pred_labelers:
                for pred_it in iterations:
                    curr_results = dict()
                    for ms in matching_schemes:
                        try:
                            curr_matches = matches.loc[ms, (tr, gt_labeler, pred_labeler, pred_it)]
                        except KeyError:
                            continue
                        if len(curr_matches) == 0:
                            continue
                        for feat in features:
                            curr_results[(ms, feat)] = peyes.match_metrics.features(
                                curr_matches, feat, verbose=False
                            )
                    if curr_results:
                        results[(tr, gt_labeler, pred_labeler, pred_it)] = curr_results
    results = pd.DataFrame.from_dict(results, orient="columns")
    results.index.names = [u.MATCHING_SCHEME_STR, peyes.constants.FEATURE_STR]
    results.columns.names = matches.columns.names
    return results


def calculate_event_sdt_measures(
        events: pd.DataFrame,
        matches: pd.DataFrame,
        pos_labels: Optional[Union[UnparsedEventLabelType, UnparsedEventLabelSequenceType]] = None,
) -> pd.DataFrame:
    results = dict()
    trials = matches.columns.get_level_values(level=peyes.constants.TRIAL_ID_STR).unique()
    gt_labelers = matches.columns.get_level_values(u.GT_STR).unique()
    pred_labelers = matches.columns.get_level_values(u.PRED_STR).unique()
    iterations = matches.columns.get_level_values(peyes.constants.ITERATION_STR).unique()
    matching_schemes = matches.index.get_level_values(u.MATCHING_SCHEME_STR).unique()
    for tr in tqdm(trials, desc="Matched Events :: SDT Metrics"):
        for gt_labeler in gt_labelers:
            try:
                trial_gt_events = events.xs((tr, gt_labeler), axis=1, level=[peyes.constants.TRIAL_ID_STR, peyes.constants.LABELER_STR])
            except KeyError:
                continue
            if trial_gt_events.size == 0:
                continue
            gt_min_iteration = np.nanmin(trial_gt_events.columns.get_level_values(peyes.constants.ITERATION_STR))
            gt_events = events[tr, gt_labeler, gt_min_iteration].dropna().values.flatten()
            if gt_events.size == 0:
                continue
            for pred_labeler in pred_labelers:
                try:
                    pred_events_all_iters = events.xs((tr, pred_labeler), axis=1, level=[peyes.constants.TRIAL_ID_STR, peyes.constants.LABELER_STR])
                except KeyError:
                    continue
                for pred_it in iterations:
                    try:
                        pred_iter_events = pred_events_all_iters[pred_it].dropna().values.flatten()
                    except KeyError:
                        continue
                    if pred_iter_events.size == 0:
                        continue
                    curr_results = dict()
                    for ms in matching_schemes:
                        try:
                            curr_matches = matches.loc[ms, (tr, gt_labeler, pred_labeler, pred_it)]
                        except KeyError:
                            continue
                        if len(curr_matches) == 0:
                            continue
                        curr_results[(ms, peyes.constants.MATCH_RATIO_STR)] = peyes.match_metrics.match_ratio(
                            pred_iter_events, curr_matches, labels=pos_labels
                        )
                        if pos_labels is not None:
                            prec, rec, f1 = peyes.match_metrics.precision_recall_f1(
                                gt_events, pred_iter_events, curr_matches, pos_labels
                            )
                            fa_rate = peyes.match_metrics.false_alarm_rate(
                                gt_events, pred_iter_events, curr_matches, pos_labels
                            )
                            d_prime, crit = peyes.match_metrics.d_prime_and_criterion(
                                gt_events, pred_iter_events, curr_matches, pos_labels
                            )
                            curr_results[(ms, peyes.constants.PRECISION_STR)] = prec
                            curr_results[(ms, peyes.constants.RECALL_STR)] = rec
                            curr_results[(ms, peyes.constants.F1_STR)] = f1
                            curr_results[(ms, peyes.constants.FALSE_ALARM_RATE_STR)] = fa_rate
                            curr_results[(ms, peyes.constants.D_PRIME_STR)] = d_prime
                            curr_results[(ms, peyes.constants.CRITERION_STR)] = crit
                    if curr_results:
                        results[(tr, gt_labeler, pred_labeler, pred_it)] = curr_results
    results = pd.DataFrame.from_dict(results, orient="columns")
    results.index.names = [u.MATCHING_SCHEME_STR, peyes.constants.METRIC_STR]
    results.columns.names = matches.columns.names
    return results



import pandas as pd

from peyes.datasets.load_dataset import load_dataset as _load_dataset
from peyes.datasets.get_metadata import get_metadata


def lund2013(directory: str = None, save: bool = False, verbose: bool = False) -> pd.DataFrame:
    """
    Loads the Lund2013 dataset from the specified directory or from the internet if the dataset is not found.

    :param directory: str; the directory to load the dataset from or save the dataset to
    :param save: bool; if True, the dataset is saved to the specified directory
    :param verbose: bool; if True, prints additional information during the loading process
    :return: a DataFrame containing the Lund2013 dataset
    """
    return _load_dataset("Lund2013", directory, save, verbose)


def irf(directory: str = None, save: bool = False, verbose: bool = False) -> pd.DataFrame:
    """
    Loads the IRF dataset from the specified directory or from the internet if the dataset is not found.

    :param directory: str; the directory to load the dataset from or save the dataset to
    :param save: bool; if True, the dataset is saved to the specified directory
    :param verbose: bool; if True, prints additional information during the loading process
    :return: a DataFrame containing the Lund2013 dataset
    """
    return _load_dataset("IRF", directory, save, verbose)


def hfc(directory: str = None, save: bool = False, verbose: bool = False) -> pd.DataFrame:
    """
    Loads the HFC dataset from the specified directory or from the internet if the dataset is not found.

    :param directory: str; the directory to load the dataset from or save the dataset to
    :param save: bool; if True, the dataset is saved to the specified directory
    :param verbose: bool; if True, prints additional information during the loading process
    :return: a DataFrame containing the Lund2013 dataset
    """
    return _load_dataset("HFC", directory, save, verbose)


def gazecom(directory: str = None, save: bool = False, verbose: bool = False) -> pd.DataFrame:
    """
    Loads the GazeCom dataset from the specified directory or from the internet if the dataset is not found.

    :param directory: str; the directory to load the dataset from or save the dataset to
    :param save: bool; if True, the dataset is saved to the specified directory
    :param verbose: bool; if True, prints additional information during the loading process
    :return: a DataFrame containing the Lund2013 dataset
    """
    return _load_dataset("GazeCom", directory, save, verbose)



from typing import Set, Dict

from tqdm import tqdm

from peyes._DataModels.Event import EventSequenceType
from peyes._DataModels.EventLabelEnum import EventLabelEnum
from peyes._DataModels.EventMatcher import EventMatcher, EventMatchesType


def match(
        ground_truth: EventSequenceType,
        prediction: EventSequenceType,
        match_by: str,
        ignore_events: Set[EventLabelEnum] = None,
        allow_xmatch: bool = False,
        **kwargs,
) -> EventMatchesType:
    """
    Match events based on the given matching criteria, ignoring specified event-labels.
    Matches can be one-to-one or one-to-many depending on the matching criteria and the specified parameters.

    :param ground_truth: a sequence of BaseEvent objects representing the ground truth events.
    :param prediction: a sequence of BaseEvent objects representing the predicted events.
    :param match_by: the matching criteria to use:
        - 'first' or 'first overlap': match the first predicted event that overlaps with each ground-truth event
        - 'last' or 'last overlap': match the last predicted event that overlaps with each ground-truth event
        - 'max' or 'max overlap': match the predicted event with maximum overlap with each ground-truth event
        - 'longest overlap': match the longest predicted event that overlaps with each ground-truth event
        - 'iou' or 'intersection over union': match the predicted event with maximum intersection-over-union
        - 'onset' or 'onset difference': match the predicted event with least onset difference
        - 'offset' or 'offset difference': match the predicted event with least offset difference
        - 'window' or 'window based': match the predicted event within a specified onset- and offset-latency window
        - 'l2' or 'l2 timing': match the predicted event with minimum timing l2 norm
        - 'other' or 'generic': match the predicted event using multiple criteria
    :param ignore_events: a set of event-labels to ignore during the matching process, default is None.
    :param allow_xmatch: if True, allows cross-matching between detectors/raters, default is False.

    :keyword min_overlap: minimum overlap required for 'first', 'last', 'max', and 'longest overlap' matching.
    :keyword min_iou: minimum intersection-over-union required for 'iou' matching.
    :keyword max_onset_difference: maximum onset difference allowed for 'onset' or 'window' matching.
    :keyword max_offset_difference: maximum offset difference allowed for 'offset' or 'window' matching.
    :keyword max_l2: maximum l2 norm allowed for 'l2' matching.

    :return: a dictionary matching each ground-truth event to event(s) from the predictions.
    """
    ignore_events = ignore_events or set()
    match_by = match_by.lower().replace("_", " ").replace("-", " ").strip()
    allow_xmatch = allow_xmatch or kwargs.pop("allow_xmatch", None) or kwargs.pop("allow_cross_match", False)
    ground_truth = [e for e in ground_truth if e.label not in ignore_events]
    prediction = [e for e in prediction if e.label not in ignore_events]
    if match_by == "first" or match_by == "first overlap":
        return EventMatcher.first_overlap(
            ground_truth, prediction, min_overlap=kwargs.pop("min_overlap", 0), allow_cross_matching=allow_xmatch
        )
    if match_by == "last" or match_by == "last overlap":
        return EventMatcher.last_overlap(
            ground_truth, prediction, min_overlap=kwargs.pop("min_overlap", 0), allow_cross_matching=allow_xmatch
        )
    if match_by == "max" or match_by == "max overlap" or match_by == "max_overlap":
        return EventMatcher.max_overlap(
            ground_truth, prediction, min_overlap=kwargs.pop("min_overlap", 0), allow_cross_matching=allow_xmatch
        )
    if "longest" in match_by and "overlap" in match_by:
        return EventMatcher.longest_overlapping_event(
            ground_truth, prediction, min_overlap=kwargs.pop("min_overlap", 0), allow_cross_matching=allow_xmatch
        )
    if match_by == "iou" or match_by == "intersection over union":
        return EventMatcher.iou(
            ground_truth, prediction, min_iou=kwargs.pop("min_iou", 0), allow_cross_matching=allow_xmatch
        )
    if match_by == "onset" or match_by == "onset difference":
        return EventMatcher.onset_difference(
            ground_truth,
            prediction,
            max_onset_difference=kwargs.pop("max_onset_difference", 0),
            allow_cross_matching=allow_xmatch
        )
    if match_by == "offset" or match_by == "offset max_onset_difference":
        return EventMatcher.offset_difference(
            ground_truth,
            prediction,
            max_offset_difference=kwargs.pop("max_offset_difference", 0),
            allow_cross_matching=allow_xmatch
        )
    if match_by == "window" or match_by == "window based":
        return EventMatcher.window_based(
            ground_truth,
            prediction,
            max_onset_difference=kwargs.pop("max_onset_difference", 0),
            max_offset_difference=kwargs.pop("max_offset_difference", 0),
            allow_cross_matching=allow_xmatch
        )
    if "l2" in match_by:
        return EventMatcher.l2_timing(
            ground_truth, prediction, max_l2=kwargs.pop("max_l2", 0), allow_cross_matching=allow_xmatch
        )
    return EventMatcher.generic_matching(ground_truth, prediction, allow_cross_matching=allow_xmatch, **kwargs)


def match_multiple(
        ground_truth: EventSequenceType,
        predictions: Dict[str, EventSequenceType],
        match_by: str,
        ignore_events: Set[EventLabelEnum] = None,
        allow_xmatch: bool = False,
        verbose: bool = False,
        **kwargs,
) -> Dict[str, EventMatchesType]:
    """
    Matched between each of the predicted event sequences and the ground-truth event sequence, using the specified
    matching criteria and ignoring specified event-labels. Matching can be one-to-one or one-to-many depending on the
    matching criteria and the specified parameters. If `verbose` is True, a progress bar tracks the matching process for
    predicted sequences.
    Returns a dictionary mapping prediction names to their respective matching results.
    See `match_events` function for more details on the matching criteria and parameters.
    """
    matches = {}
    for name, pred in tqdm(predictions.items(), desc="Matching", disable=not verbose):
        matches[name] = match(
            ground_truth, pred, match_by, ignore_events=ignore_events, allow_xmatch=allow_xmatch, **kwargs
        )
    return matches




import copy
from typing import List

import numpy as np


def is_one_dimensional(arr) -> bool:
    """ Returns true if the array's shape is (n,) or (1, n) or (n, 1) """
    arr = np.asarray(arr)
    if arr.ndim == 1:
        return True
    if arr.ndim == 2 and min(arr.shape) == 1:
        return True
    return False


def pair_boolean_arrays(arr1: np.ndarray, arr2: np.ndarray) -> np.ndarray:
    """
    Given two boolean arrays, match `True` values between the two arrays such that the indices of the matched values
    are as close as possible. Returns a 2D array where each row is a matching pair of indices, where the first column
    is the index in `arr1` and the second column is the index in `arr2`.
    See also: https://stackoverflow.com/q/78484847/8543025

    :param arr1: boolean array
    :param arr2: boolean array
    :return: array of integer pairs (matched indices), with shape m×2 (0 <= m <= min(sum(arr1), sum(arr2))

    :raises ValueError: if the input arrays are not one-dimensional
    """
    if not is_one_dimensional(arr1) or not is_one_dimensional(arr2):
        raise ValueError("input arrays must be one-dimensional")
    arr1 = np.asarray(arr1).reshape(-1).astype(bool)
    arr2 = np.asarray(arr2).reshape(-1).astype(bool)
    arr1_idxs, arr2_idxs = np.where(arr1)[0], np.where(arr2)[0]
    diffs = abs(arr1_idxs[:, None] - arr2_idxs[None, :])
    if diffs.size == 0:
        return np.empty((0, 2), dtype=int)
    rowwise_argmin = np.stack([diffs.argmin(0), np.arange(diffs.shape[1])]).T
    colwise_argmin = np.stack([np.arange(diffs.shape[0]), diffs.argmin(1)]).T
    is_matching = (rowwise_argmin[:, None] == colwise_argmin).all(-1).any(1)
    idxs = rowwise_argmin[is_matching]
    matching_indices = np.stack([arr1_idxs[idxs[:, 0]], arr2_idxs[idxs[:, 1]]]).T
    return matching_indices


def get_chunk_indices(arr) -> List[np.ndarray]:
    """
    Given a 1D array with multiple values, returns a list of arrays, where each array contains the indices of
    a different "chunk", i.e. a sequence of the same value.
    """
    if not is_one_dimensional(arr):
        raise ValueError("arr must be one-dimensional")
    indices = np.arange(len(arr))
    split_on = np.nonzero(np.diff(arr))[0] + 1  # +1 because we want to include the last index of each chunk
    chunk_indices = np.split(indices, split_on)
    return chunk_indices


def merge_chunks(arr: np.ndarray, max_samples_between: int) -> np.ndarray:
    """
    Splits the input array into chunks of identical values. If two chunks of the same value are separated by a chunk
    of a different value, and the middle chunk is shorter/equal to `max_samples_between`, it is set to the same value as
    the other two chunks.
    Returns a new array with the merged chunks.
    """
    if not is_one_dimensional(arr):
        raise ValueError("input array must be one-dimensional")
    if max_samples_between < 0:
        raise ValueError("max_samples_between must be non-negative")
    arr_copy = copy.deepcopy(arr)
    chunk_indices = get_chunk_indices(arr_copy)
    for i, middle_chunk in enumerate(chunk_indices[1:-1], start=1):
        if len(middle_chunk) > max_samples_between:
            # skip if the chunk is too long
            continue
        left_chunk_val = arr_copy[chunk_indices[i - 1][0]]
        right_chunk_val = arr_copy[chunk_indices[i + 1][0]]
        if left_chunk_val != right_chunk_val:
            # skip if the left and right chunks have different values
            continue
        arr_copy[middle_chunk] = left_chunk_val
    return arr_copy


def reset_short_chunks(arr: np.ndarray, min_samples: int, default_value: int) -> np.ndarray:
    """
    Splits the input array into chunks of identical values. If a chunk is shorter than `min_samples`, it is set to the
    `default_value`. Returns a new array with the reset chunks.
    """
    if not is_one_dimensional(arr):
        raise ValueError("input array must be one-dimensional")
    if min_samples < 0:
        raise ValueError("min_samples must be non-negative")
    arr_copy = copy.deepcopy(arr)
    chunk_indices = get_chunk_indices(arr_copy)
    for chunk in chunk_indices:
        if len(chunk) < min_samples:
            arr_copy[chunk] = default_value
    return arr_copy




import os
from typing import List, Optional, Union, Tuple, Sequence, Callable

import numpy as np
import pandas as pd
import scipy.stats as stats
import scikit_posthocs as sp
import plotly.graph_objects as go

import peyes

import analysis.utils as u
from peyes._utils.visualization_utils import make_empty_figure
from peyes._DataModels.UnparsedEventLabel import UnparsedEventLabelType, UnparsedEventLabelSequenceType


def extract_subframe(
        data: pd.DataFrame,
        level: Union[str, int],
        value: Union[str, int, Sequence[str], Sequence[int]],
        axis: int = 0,
        drop_single_values: bool = True,  # drop level if only one value is selected
) -> pd.DataFrame:
    if isinstance(value, str) or isinstance(value, int):
        value = [value]
    if axis == 0:
        is_values = data.index.get_level_values(level).isin(value)
        subframe = data.loc[is_values]
    elif axis == 1:
        is_values = data.columns.get_level_values(level).isin(value)
        subframe = data.loc[:, is_values]
    else:
        raise ValueError("`axis` must be 0 or 1.")
    if drop_single_values and len(value) == 1:
        subframe = subframe.droplevel(level, axis=axis)
    return subframe


def load_data(
        dataset_name: str,
        output_dir: str,
        data_dir_name: str,
        label: Optional[Union[UnparsedEventLabelType, UnparsedEventLabelSequenceType]] = None,
        filename_prefix: str = "",
        filename_suffix: str = "",
        iteration: Optional[int] = 1,
        stimulus_type: Optional[Union[str, List[str]]] = None,
        sub_index: Optional[Union[str, List[str]]] = None,
) -> pd.DataFrame:
    data_dir_path = os.path.join(output_dir, dataset_name, data_dir_name)
    fullpath = os.path.join(data_dir_path, u.get_filename_for_labels(
        label, prefix=filename_prefix, suffix=filename_suffix, extension="pkl"
    ))
    try:
        data = pd.read_pickle(fullpath)
    except FileNotFoundError:
        raise FileNotFoundError(f"Couldn't find `{fullpath}`. Please preprocess the dataset first.")
    if iteration is not None:
        data = data.xs(iteration, level=peyes.constants.ITERATION_STR, axis=1, drop_level=True)
    if stimulus_type:
        stim_trial_ids = u.get_trials_for_stimulus_type(dataset_name, stimulus_type)
        is_trial_ids = data.columns.get_level_values(peyes.constants.TRIAL_ID_STR).isin(stim_trial_ids)
        data = data.loc[:, is_trial_ids]
    if sub_index:
        data = extract_subframe(data, level=0, value=sub_index, axis=0, drop_single_values=False)
    return data


def mann_whitney(
        data: pd.DataFrame,
        gt_cols: Union[str, Sequence[str]],
        alternative: str = "two-sided",
        method: str = "auto",
) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """
    Run the Mann-Whitney U-test for each (metric, GT labeler) pair.
    Mann-Whitney is the non-parametric equivalent of two-sampled t-test (using ranks instead of means), which assumes
    independent measures between groups.

    Returns the U-statistic, p-value and number of samples for each pair.
    Read the docstring for `_statistical_analysis` for implementation details.
    """
    pred = data.columns.get_level_values(u.PRED_STR).unique()
    if len(pred) > 2:
        raise ValueError(
            f"Mann-Whitney's U-test requires exactly 2 predictors, got {len(pred)}. " +
            "Consider using Kruskal-Wallis test instead."
        )
    mw_test = lambda vals: stats.mannwhitneyu(*vals, alternative=alternative, method=method, nan_policy='omit')
    u_stats, p_vals, _, Ns = _statistical_analysis(data, gt_cols, mw_test)
    return u_stats, p_vals, Ns


def wilcoxon(data: pd.DataFrame,
             gt_cols: Union[str, Sequence[str]],
             alternative: str = "two-sided",
             method: str = "auto",
             zero_method: str = 'wilcox',
) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """
    Run the Wilcoxon signed-rank for each (metric, GT labeler) pair.
    Wilcoxon is the non-parametric equivalent of paired-samples t-test (using ranks instead of means), which assumes
    repeated measures between groups.

    Returns the W-statistic, p-value and number of samples for each pair.
    Read the docstring for `_statistical_analysis` for implementation details.
    """
    def wilcoxon_test(vals):
        assert 0 < len(vals) <=2, f"Wilcoxon test can only run on 1 set of differences or 2 sets of samples, {len(vals)} given."
        if len(vals) == 1:
            return stats.wilcoxon(
                vals, alternative=alternative, method=method, zero_method=zero_method, nan_policy='omit'
            )
        vals1, vals2 = vals[0], vals[1]
        assert len(vals1) == len(vals2), f"length mismatch: |group 1|={len(vals1)} and |group 2|={len(vals2)}"
        return stats.wilcoxon(
            x=vals1, y=vals2, alternative=alternative, method=method, zero_method=zero_method, nan_policy='omit'
        )

    w_stats, p_vals, _, Ns = _statistical_analysis(data, gt_cols, wilcoxon_test)
    return w_stats, p_vals, Ns


def kruskal_wallis_dunns(
        data: pd.DataFrame,
        gt_cols: Union[str, Sequence[str]],
        multi_comp: Optional[str] = "fdr_bh",
) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """
    Run the Kruskal-Wallis test with Dunn's post-hoc test for multiple comparisons, for each (metric, GT labeler) pair.
    Kruskal-Wallis is the non-parametric equivalent to one-way ANOVA (using ranks instead of means), which assumes
    independent samples between groups. Dunn's test is the common post-hoc test used after a significant KW, to
    determine which pairs of groups are significantly different.

    Returns the KW-statistic, KW-p-value, Dunn's-p-values and number of samples for each pair.
    Read the docstring for `_statistical_analysis` for implementation details.
    """
    kw_test = lambda vals: stats.kruskal(*vals, nan_policy='omit')
    dunns_test = lambda vals: sp.posthoc_dunn(a=list(vals), p_adjust=multi_comp)
    h_stat, p_vals, dunn_p_vals, Ns = _statistical_analysis(data, gt_cols, kw_test, dunns_test)
    return h_stat, p_vals, dunn_p_vals, Ns


def friedman_nemenyi(
        data: pd.DataFrame,
        gt_cols: Union[str, Sequence[str]],
) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """
    Run the Friedman test with Nemenyi's post-hoc test for multiple comparisons, for each (metric, GT labeler) pair.
    Friedman is the non-parametric equivalent to repeated measures ANOVA (using ranks instead of means), which assumes
    paired samples between groups. Nemenyi's test is the common post-hoc test used after a significant Friedman, to
    determine which pairs of groups are significantly different.
    Note Nemenyi's test doesn't require correcting for multiple comparisons.

    Returns the Friedman's Fr-statistic, p-value, Nemenyi's p-values and number of samples for each pair.
    Read the docstring for `_statistical_analysis` for implementation details.
    """
    friedman_test = lambda vals: stats.friedmanchisquare(*vals, nan_policy='omit')
    nemenyi_test = lambda vals: sp.posthoc_nemenyi(a=list(vals))
    q_stat, p_vals, nemenyi_p_vals, Ns = _statistical_analysis(data, gt_cols, friedman_test, nemenyi_test)
    return q_stat, p_vals, nemenyi_p_vals, Ns


def create_post_hoc_table(
        post_hoc_results: pd.DataFrame,
        metric: str,
        *gt_names: str,
        alpha: float = 0.05,
        marginal_alpha: Optional[float] = 0.075,
) -> (pd.DataFrame, pd.DataFrame):
    assert 0 < alpha < 1, f"parameter `alpha` must be in range (0, 1), {alpha: .3f} given."
    assert marginal_alpha is None or alpha < marginal_alpha < 1, f"parameter `marginal_alpha` must be in range ({alpha: .3f}, 1), {marginal_alpha: .3f} given."
    pvals = pd.concat(
        [post_hoc_results.loc[metric, gt] for gt in gt_names], keys=gt_names
    ).reorder_levels([1, 0])
    pvals.sort_index(
        level=0, key=lambda idx: idx.map(lambda det: list(pvals.columns).index(det)), inplace=True
    )
    pvals.index.names = [u.PRED_STR, u.GT_STR]
    pvals.columns.names = [u.PRED_STR]

    table = np.full_like(pvals, "n.s.", dtype=np.dtypes.StringDType())
    if marginal_alpha is not None:
        table[pvals <= marginal_alpha] = '†'
    table[pvals <= alpha] = '*'
    table[pvals <= alpha / 5] = '**'
    table[pvals <= alpha / 50] = '***'
    table = pd.DataFrame(table, index=pvals.index, columns=pvals.columns)

    for c, det1 in enumerate(pvals.index.get_level_values(0).unique()):
        for r, det2 in enumerate(pvals.index.get_level_values(0).unique()):
            if c == r:
                table.loc[det1, det2] = "--"
            elif r > c:
                table.loc[det2, det1] = ""
            else:
                for gt in gt_names:
                    table.loc[(det1, gt), det2] = f"{pvals.loc[(det1, gt), det2] :.4f}"
    return table



def distributions_figure(
        data: pd.DataFrame,
        gt1: str,
        title: str,
        colors: u.COLORMAP_TYPE = None,
        gt2: Optional[str] = None,
        only_box: bool = False,
        show_other_gt: bool = False,
        share_x: bool = False,
        share_y: bool = False,
        subplots_vspace: float = 0.1,
        subplots_hspace: float = 0.1,
) -> go.Figure:
    """
    Creates a violin/box subplot for each unique value in the input DataFrame's index. Each subplot (index value)
    contains violins/boxes for all predictor labelers (detection algorithms), compared to the specified GT labeler(s).
    If two GT labelers are provided, the plot will show GT1 in the positive (right) side and GT2 in the negative (left) side.

    :param data: DataFrame; Should have the following MultiIndex structure:
        - Index :: single level
        - Columns :: 1st level: trial id
        - Columns :: 2nd level: GT labeler
        - Columns :: 3rd level: detector
    :param gt1: name of the first GT labeler to compare.
    :param title: optional; title for the plot.
    :param colors: optional; list of hex colors to mark different labelers in the violins/boxes.
    :param gt2: optional; name of the second GT labeler to compare.
    :param only_box: if True, only box plots will be shown.
    :param show_other_gt: if True, a separate violin/box plot will be shown for the other GT labeler.
    :param share_x: if True, subplots will share x-axis within each column.
    :param share_y: if True, subplots will share y-axis within each row.
    :param subplots_vspace: vertical spacing between subplots.
    :param subplots_hspace: horizontal spacing between subplots.

    :return: Plotly figure with the violin/box plot.
    """
    gt_cols = list(filter(None, [gt1, gt2]))
    assert 0 < len(gt_cols) <= 2
    indices = sorted(
        data.index.unique(),
        key=lambda met: u.METRICS_CONFIG[met][1] if met in u.METRICS_CONFIG else ord(met[0])
    )
    fig, nrows, ncols = make_empty_figure(
        subtitles=list(map(lambda idx: u.METRICS_CONFIG[idx][0] if idx in u.METRICS_CONFIG else idx, indices)),
        sharex=share_x, sharey=share_y,
        subplots_vspace=subplots_vspace, subplots_hspace=subplots_hspace,
    )
    for i, idx in enumerate(indices):
        r, c = (i, 0) if ncols == 1 else divmod(i, ncols)
        for j, gt_col in enumerate(gt_cols):
            violin_side = None if len(gt_cols) == 1 else 'positive' if j == 0 else 'negative'
            opacity = 0.75 if len(gt_cols) == 1 else 0.75 if j == 0 else 0.25
            gt_df = (data.xs(gt_col, level=u.GT_STR, axis=1).loc[idx]).unstack()
            detectors = u.sort_labelers(gt_df.columns.get_level_values(u.PRED_STR).unique())
            for k, det in enumerate(detectors):
                if det not in gt_cols:
                    det_name = det.removesuffix("Detector")
                    det_color = u.get_labeler_color(det_name, k, colors)
                elif det in gt_cols and show_other_gt:
                    det_name = "Other GT"
                    det_color = "#bab0ac"
                else:
                    continue
                if only_box:
                    fig.add_trace(
                        row=r + 1, col=c + 1,
                        trace=go.Box(
                            x0=det_name, y=gt_df[det].explode().dropna().values,
                            name=f"{gt_col}, {det_name}", legendgroup=det_name,
                            marker_color=det_color, line_color=det_color,
                            opacity=opacity, boxmean='sd', showlegend=i == 0,
                        )
                    )
                else:
                    fig.add_trace(
                        row=r + 1, col=c + 1,
                        trace=go.Violin(
                            x0=det_name, y=gt_df[det].explode().dropna().values,
                            side=violin_side, opacity=opacity, spanmode='hard',
                            fillcolor=det_color, line_color='black',
                            name=f"{gt_col}, {det_name}", legendgroup=det_name, scalegroup=idx, showlegend=i == 0,
                            box_visible=True, meanline_visible=True,
                        ),
                    )
        y_range = u.METRICS_CONFIG[idx][2] if idx in u.METRICS_CONFIG else None
        fig.update_yaxes(row=r + 1, col=c + 1, range=y_range)
    fig.update_layout(
        title=title,
        violinmode='overlay',
        boxmode='group',
        boxgroupgap=0,
        boxgap=0,
    )
    return fig


def calc_error_bars(sub_data: pd.DataFrame, error_bars: Optional[str] = None) -> pd.Series:
    error_bars = error_bars.strip().lower() if error_bars else None
    if error_bars is None:
        return None
    if error_bars == 'ci':
        # see https://stackoverflow.com/a/69841368/8543025
        ci = sub_data.apply(    # symmetric CI around 0
            lambda x: stats.t.interval(0.95, len(x) - 1, loc=0, scale=stats.sem(x)), axis=1
        )
        return ci.apply(lambda x: x[1])     # only the top of the CI
    std = sub_data.std(axis=1)
    if error_bars == "std":
        return std
    elif error_bars == "sem":
        return std / np.sqrt(sub_data.count(axis=1))
    raise ValueError(f"Invalid error bars type: {error_bars}, choose from: sem, std, ci")


def _statistical_analysis(
        data: pd.DataFrame,
        gt_cols: Union[str, Sequence[str]],
        test: Callable[[Sequence[Sequence[float]]], Tuple[float, float]],
        post_hoc_test: Optional[Callable[[Sequence[Sequence[float]]], pd.DataFrame]] = None,
) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """
    Runs the statistical test and post-hoc test for each (metric, GT labeler) pair in the input DataFrame.
    For each unique value in the input DataFrame's index (metric names) and each of the GT labelers, we:
    1. Extract the values for each predictor labeler (detection algorithm) over all trials, so that we have a list of
        values for each predictor.
    2. Run the statistical test with the extracted values.
    3. If a post-hoc test is provided, we run it with the extracted values.
    4. Store the results in the output DataFrames.

    Raises a ValueError if there are less than 2 predictors.
    Returns four DataFrames containing results for the statistical analysis:
    - statistics: index is metric name, columns are GT labelers, values are the test statistic (float).
    - pvalues: index is metric name, columns are GT labelers, values are the test p-value (float).
    - post_hoc_res: index is metric name, columns are GT labelers, values are the post-hoc test results: a DataFrame
        with predictors as index and columns, and p-values as values. If no post-hoc test is provided, this DataFrame
        will be empty.
    - Ns: index is metric name, columns are multiindex with pairs of (GT, Pred) labelers, values are the number of
        measurements (trials) for each pair.

    :param data: DataFrame; Should have the following MultiIndex structure:
        - Index :: single level metric names
        - Columns :: 1st level: trial id
        - Columns :: 2nd level: GT labeler
        - Columns :: 3rd level: Pred labeler (detection algorithm)
    :param gt_cols: GT labeler(s) to compare against predictors (detection algorithms).
    :param test: Statistical test to run with the extracted values.
    :param post_hoc_test: Optional; post-hoc test to run with the extracted values.

    :return: Four DataFrames - statistics, pvalues, post_hoc_res, Ns.
    """
    predictors = data.columns.get_level_values(u.PRED_STR).unique()
    if len(predictors) < 2:
        raise ValueError(f"Not enough predictors for a statistical test: {len(predictors)}")
    if len(predictors) == 2 and post_hoc_test is not None:
        raise RuntimeError("You don't need a post-hoc test for only 2 predictors.")
    gt_cols = gt_cols if isinstance(gt_cols, list) else [gt_cols]
    metrics = sorted(
        data.index.unique(),
        key=lambda met: u.METRICS_CONFIG[met][1] if met in u.METRICS_CONFIG else ord(met[0])
    )
    Ns = {}                                         # (metric, GT, Pred) -> num_trials
    statistics, pvalues, post_hoc_res = {}, {}, {}      # (metric, GT) -> value
    for _i, gt_col in enumerate(gt_cols):
        for _j, met in enumerate(metrics):
            gt_series = data.xs(gt_col, level=u.GT_STR, axis=1).loc[met]
            if pd.isna(gt_series).all():
                continue
            gt_df = gt_series.unstack().drop(columns=gt_cols, errors='ignore')  # drop other GT labelers

            # extract per-detector values
            detectors = u.sort_labelers(gt_df.columns.get_level_values(u.PRED_STR).unique())
            detector_values = {}
            for det in detectors:
                vals = gt_df[det].explode().values.astype(float)
                n = sum(~np.isnan(vals))
                Ns[(met, gt_col, det)] = n
                if n > 0:
                    detector_values[det] = vals

            # calculate statistical test
            res = test(list(detector_values.values()))
            statistic, pvalue = res.statistic, res.pvalue
            statistics[(met, gt_col)] = statistic
            pvalues[(met, gt_col)] = pvalue

            # calculate post-hoc test
            if post_hoc_test is not None:
                post_result = pd.DataFrame(
                    post_hoc_test(list(detector_values.values())).values,
                    index=list(detector_values.keys()), columns=list(detector_values.keys())
                )
            else:
                post_result = pd.DataFrame(np.array([]))
            post_result.index.names = post_result.columns.names = [u.PRED_STR]
            post_hoc_res[(met, gt_col)] = post_result

    # create outputs
    statistics = pd.Series(statistics).unstack()
    pvalues = pd.Series(pvalues).unstack()
    post_hoc_res = pd.Series(post_hoc_res).unstack()
    statistics.index.names = pvalues.index.names = post_hoc_res.index.names = [peyes.constants.METRIC_STR]
    statistics.columns.names = pvalues.columns.names = post_hoc_res.columns.names = [u.GT_STR]
    Ns = pd.Series(Ns)
    Ns.index.names = [peyes.constants.METRIC_STR, u.GT_STR, u.PRED_STR]
    Ns = Ns.unstack([u.GT_STR, u.PRED_STR])
    return statistics, pvalues, post_hoc_res, Ns



import warnings
from itertools import islice
from typing import Sequence, Optional

import numpy as np
import pandas as pd
import Levenshtein
from scipy.stats import norm


def transition_matrix(seq: Sequence, normalize_rows: bool = False) -> pd.DataFrame:
    """
    Calculates the transition matrix of a sequence of events. The matrix is a DataFrame where the rows represent the
    "from" event and the columns represent the "to" event. The values in the matrix are the counts of transitions
    between the events. If `normalize_rows` is True, the values in each row are normalized to sum to 1.
    See implementation details at https://stackoverflow.com/a/47298184/8543025.
    """
    def window(s, n=2):
        """ Sliding window width n from sequence s """
        it = iter(s)
        result = tuple(islice(it, n))
        if len(result) == n:
            yield result
        for elem in it:
            result = result[1:] + (elem,)
            yield result
    pairs = pd.DataFrame(window(seq), columns=['From', 'To'])
    counts = pairs.groupby('From')['To'].value_counts().unstack(fill_value=0)
    if normalize_rows:
        counts = counts.div(counts.sum(axis=1), axis=0)
    return counts


def normalized_levenshtein_distance(gt: Sequence, pred: Sequence) -> float:
    """
    Calculates the normalized Levenshtein distance between two sequences. Normalization is with respect to the size of
    the GT, which may result in a value exceeding 1 if the predicted sequence is longer than the GT sequence.
    This is the same as the Word Error Rate (WER) used in speech recognition.
    For more information on using 1-NLD in eye-tracking, see https://doi.org/10.3758/s13428-021-01763-7
    """
    d = Levenshtein.distance(gt, pred)
    normalized_d = d / len(gt)
    return normalized_d


def dprime_and_criterion(p: int, n: float, pp: int, tp: int, correction: Optional[str]) -> (float, float):
    """
    Calculates d-prime and criterion while optionally applying a correction for floor/ceiling effects on the hit-rate
    and/or false-alarm rate. See information on correction methods at https://stats.stackexchange.com/a/134802/288290.
    See implementation details at https://lindeloev.net/calculating-d-in-python-and-php/.

    :param p: int; number of positive GT events
    :param n: int; number of negative GT events
    :param pp: int; number of positive predicted events
    :param tp: int; number of true positive predictions
    :param correction: str; optional correction method for floor/ceiling effects
    :return:
        - d_prime: float; the d-prime value
        - criterion: float; the criterion value
    """
    hr, far = _dprime_rates(p, n, pp, tp, correction)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", RuntimeWarning)
        d_prime = norm.ppf(hr) - norm.ppf(far)
        criterion = -0.5 * (norm.ppf(hr) + norm.ppf(far))
        return d_prime, criterion


def _dprime_rates(p: int, n: float, pp: int, tp: int, correction: Optional[str]) -> (float, float):
    """
    Calculates hit-rate and false-alarm rate for computing d-prime. Optionally applies a correction for floor/ceiling
    effects on the rates. See information on correction methods at https://stats.stackexchange.com/a/134802/288290.
    Returns a tuple of (hit-rate, false-alarm rate).
    """
    fp = pp - tp
    assert 0 <= tp <= min(p, pp), f"True Positive count must be between 0 and min(p, pp) = {min(p, pp)}; got: {tp}"
    # assert 0 <= fp <= n, f"False Positive count must be between 0 and n = {n}; got: {fp}"
    # NOTE: when comparing events (not labels), it doesn't hold that 0 <= fp <= n
    hit_rate = tp / p if p > 0 else np.nan
    false_alarm_rate = fp / n if n > 0 else np.nan
    if 0 < hit_rate < 1 and 0 < false_alarm_rate < 1:
        # no correction needed
        return hit_rate, false_alarm_rate
    corr = (correction or "").lower().strip().replace(" ", "_").replace("-", "_")
    if corr is None or not corr:
        return hit_rate, false_alarm_rate
    if corr in {"mk", "m&k", "macmillan_kaplan", "macmillan"}:
        # apply Macmillan & Kaplan (1985) correction
        if hit_rate == 0:
            hit_rate = 0.5 / p
        if hit_rate == 1:
            hit_rate = 1 - 0.5 / p
        if false_alarm_rate == 0:
            false_alarm_rate = 0.5 / n
        if false_alarm_rate == 1:
            false_alarm_rate = 1 - 0.5 / n
        return hit_rate, false_alarm_rate
    if correction in {"ll", "loglinear", "log_linear", "hautus"}:
        # apply Hautus (1995) correction
        prevalence = p / (p + n) if p + n > 0 else np.nan
        new_tp, new_fp = tp + prevalence, fp + 1 - prevalence
        new_p, new_n = p + 2 * prevalence, n + 2 * (1 - prevalence)
        hit_rate = new_tp / new_p if new_p > 0 else np.nan
        false_alarm_rate = new_fp / new_n if new_n > 0 else np.nan
        return hit_rate, false_alarm_rate
    raise ValueError(f"Invalid correction: {correction}")

